/***************************************************************************

  draw.c

  The SDL component

  Draw management

  (c) 2004-2005 Laurent Carlier <lordheavy@infonie.fr>
                Benot Minisini <gambas@users.sourceforge.net>
                Bodard Fabien   <gambasfr@wanadoo.fr>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

#define __DRAW_C

#include "main.h"
#include "gambas.h"

#include "draw.h"
#include "surface.h"
#include "font.h"

#include "SDL.h"
#include "SFont.h"
#include "SDL_ttf.h"
#include "SDL_gfxPrimitives.h"

#define DRAW_STACK_MAX 8

static CDRAWING draw_stack[DRAW_STACK_MAX];
static CDRAWING *draw_current = 0;

#define THIS    (draw_current)
#define SURF    (THIS->Surf)
#define CCOLOR  (THIS->CColor)
#define LASTX   (THIS->LastX)
#define LASTY   (THIS->LastY)
#define FONT    (THIS->font)

/**************************************************************************

  Draw Class - init/exit methods

***************************************************************************/

BEGIN_METHOD(DRAWING_begin, GB_OBJECT surface)

  if (THIS >= &draw_stack[DRAW_STACK_MAX - 1])
  {
    GB.Error("Too many nested drawings");
    return;
  }

  if (THIS == 0)
    THIS = draw_stack;
  else
    THIS++;

  CCOLOR = 0xFF;
  LASTX = LASTY = 0;
  FONT = NULL;

  if (MISSING(surface))
  {
    if (!MainScreen)
    {
      GB.Error("Screen not defined !");
    }

    SURF = 0;
    return;
  }
  else
  {
    void *surface = VARG(surface);

    if (GB.CheckObject(surface))
      return;

    GB.StoreObject(ARG(surface), (void **) &SURF);
  }

END_METHOD

BEGIN_METHOD_VOID(DRAWING_end)

  if (!THIS)
    return;

  if (SURF)
    GB.Unref((void **) &SURF);

  if (FONT)
    GB.Unref((void **) &FONT);

  if (THIS == draw_stack)
    THIS = 0;
  else
    THIS--;

END_METHOD

/**************************************************************************

  Point - for drawing pixel on the screen or surfaces

***************************************************************************/

BEGIN_METHOD(DRAWING_point, GB_INTEGER X; GB_INTEGER Y)

  SDL_Surface *DrawSurf = MainScreen;
  LASTX = VARG(X);
  LASTY = VARG(Y);

  if (!THIS)
  {
    GB.Error("Draw.Begin() missing !");
    return;
  }

  if (SURF)
    DrawSurf = GETSURFACE(SURF);

  pixelColor(DrawSurf, LASTX, LASTY, CCOLOR);

END_METHOD

/**************************************************************************

  Line   - draw a line from a point to another
  LineTo - draw a line from the last point to another relative point

***************************************************************************/

BEGIN_METHOD(DRAWING_line, GB_INTEGER X1; GB_INTEGER Y1; GB_INTEGER X2; GB_INTEGER Y2)

  Sint16 x1, y1;
  SDL_Surface *DrawSurf = MainScreen;

  x1 = VARG(X1);
  y1 = VARG(Y1);

  LASTX = VARG(X2);
  LASTY = VARG(Y2);

  if (!THIS)
  {
    GB.Error("Draw.Begin() missing !");
    return;
  }

  if (SURF)
  {
    DrawSurf = GETSURFACE(SURF);
  }

  lineColor(DrawSurf, x1, y1, LASTX, LASTY, CCOLOR);

END_METHOD

BEGIN_METHOD(DRAWING_lineto, GB_INTEGER X; GB_INTEGER Y)

  Sint16 x1, y1;
  SDL_Surface *DrawSurf = MainScreen;

  x1 = LASTX;
  y1 = LASTY;

  LASTX = x1 + VARG(X);
  LASTY = y1 + VARG(Y);

  if (!THIS)
  {
    GB.Error("Draw.Begin() missing !");
    return;
  }

  if (SURF)
    DrawSurf = GETSURFACE(SURF);

  lineColor(DrawSurf, x1, y1, LASTX, LASTY, CCOLOR);

END_METHOD

/**************************************************************************

  Rect  - draw an zmpty rectangle
          draw a filled rectangle if filled is TRUE

***************************************************************************/

BEGIN_METHOD(DRAWING_rect, GB_INTEGER X; GB_INTEGER Y; GB_INTEGER width; GB_INTEGER height)

  Sint16 x, y;
  SDL_Surface *DrawSurf = MainScreen;

  x = VARG(X);
  y = VARG(Y);

  LASTX = x + VARG(width)-1;
  LASTY = y + VARG(height)-1;

  if (!THIS)
  {
    GB.Error("Draw.Begin() missing !");
    return;
  }

  if (SURF)
  {
    DrawSurf = GETSURFACE(SURF);
  }

  if (THIS->filled)
    boxColor(DrawSurf, x, y, LASTX, LASTY, CCOLOR);
  else
    rectangleColor(DrawSurf, x, y, LASTX, LASTY, CCOLOR);

END_METHOD

/**************************************************************************

  Ellipse  - draw an zmpty ellipse
             draw a filled ellipse if filled is TRUE

***************************************************************************/

BEGIN_METHOD(DRAWING_ellipse, GB_INTEGER X; GB_INTEGER Y; GB_INTEGER width; GB_INTEGER height)

  Sint16 x, y, rx, ry;
  SDL_Surface *DrawSurf = MainScreen;

  x = VARG(X);
  y = VARG(Y);
  rx = VARG(width);
  ry = VARG(height);

  if (!THIS)
  {
    GB.Error("Draw.Begin() missing !");
    return;
  }

  if (SURF)
  {
    DrawSurf = GETSURFACE(SURF);
  }

  if (THIS->filled)
    filledEllipseColor(DrawSurf, x, y, rx, ry, CCOLOR);
  else
    ellipseColor(DrawSurf, x, y, rx, ry, CCOLOR);

END_METHOD

/**************************************************************************

  Font - Set/Get used font

***************************************************************************/

BEGIN_PROPERTY(DRAWING_font)

  if (READ_PROPERTY)
  {
     if (!FONT)
       return;
  }
  else
  {
    void *font = VPROP(GB_OBJECT);

    if (GB.CheckObject(font))
      return;

    GB.StoreObject(PROP(GB_OBJECT), (void **) &FONT);
  }

END_PROPERTY

/**************************************************************************

  Color - Set/Get used color (RRGGBBAA)

***************************************************************************/

BEGIN_PROPERTY(DRAWING_color)

  if (READ_PROPERTY)
    GB.ReturnInteger(CCOLOR);
  else
    CCOLOR = VPROP(GB_INTEGER);

END_PROPERTY

/**************************************************************************

  Text - Print text with font

***************************************************************************/

BEGIN_METHOD(DRAWING_text, GB_STRING text; GB_INTEGER X; GB_INTEGER Y;)

  Sint16 x, y;
  CFONT *font = FONT;
  SDL_Surface *DrawSurf = MainScreen;
  SDL_Rect rect;

  x = VARGOPT(X, 0);
  y = VARGOPT(Y, 0);

  if (!font)
    return;

  if (!THIS)
  {
    GB.Error("Draw.Begin() missing !");
    return;
  }

  if (SURF)
    DrawSurf = GETSURFACE(SURF);

  if (font->fontType == FONT_SFONT)
  {
    if MISSING(X)
      x = (DrawSurf->w / 2) - (SFont_TextWidth(font->id, STRING(text))/2);
    
    if MISSING(Y)
      y = (DrawSurf->h / 2) - (SFont_TextHeight(font->id)/2);
    
    SFont_Write (DrawSurf, font->id, x, y, STRING(text));
  }

  if (font->fontType == FONT_TTF)
  {
    SDL_Color color;
    SDL_Surface *surface;
    int w,h = 0;
    
    if MISSING(X)
    {
      TTF_SizeUTF8(font->id, STRING(text), &w, &h);
      x = (DrawSurf->w / 2) - (w/2);
    }
    
    if MISSING(Y)
    {
      TTF_SizeUTF8(font->id, STRING(text), &w, &h);
      y = (DrawSurf->h / 2) - (h/2);;
    }
    
    color.r = ((CCOLOR >> 24) & 0xFF);
    color.g = ((CCOLOR >> 16) & 0xFF);
    color.b = ((CCOLOR >> 8)  & 0xFF);
    if (font->quality)
      surface = TTF_RenderUTF8_Blended (font->id, STRING(text), color);
    else
      surface = TTF_RenderUTF8_Solid (font->id, STRING(text), color);
   
    if (!surface)
    {
      GB.Error("Cannot print the text : &1",TTF_GetError());
      return;
    }
      
    rect.x = x;
    rect.y = y;
    SDL_BlitSurface(surface, NULL, DrawSurf, &rect);
    SDL_FreeSurface(surface);
  }
 
END_METHOD

/**************************************************************************

  Fill - Define if shapes are filled

***************************************************************************/

BEGIN_PROPERTY(DRAWING_fill)

  if (READ_PROPERTY)
    GB.ReturnBoolean(THIS->filled);
  else
  {
    if (VPROP(GB_BOOLEAN))
      THIS->filled = 1;
    else
      THIS->filled = 0;
  }

END_PROPERTY

/**************************************************************************/

GB_DESC Cdrawing[] =
{
  GB_DECLARE("Draw", 0), GB_NOT_CREATABLE(),

  GB_STATIC_METHOD("Begin", NULL, DRAWING_begin, "[(surface)Surface;]"),
  GB_STATIC_METHOD("End",   NULL, DRAWING_end,   NULL),

  GB_STATIC_METHOD("Point",    NULL, DRAWING_point,    "(X)i(Y)i"),
  GB_STATIC_METHOD("Line",     NULL, DRAWING_line,     "(X1)i(Y1)i(X2)i(Y2)i"),
  GB_STATIC_METHOD("LineTo",   NULL, DRAWING_lineto,   "(X)i(Y)i"),
  GB_STATIC_METHOD("Rect",     NULL, DRAWING_rect,     "(X)i(Y)i(Width)i(Height)i"),
  GB_STATIC_METHOD("Ellipse",  NULL, DRAWING_ellipse,  "(X)i(Y)i(Width)i(Height)i"),
  GB_STATIC_METHOD("Text",     NULL, DRAWING_text,     "(text)s[(X)i(Y)i]"),
  
  GB_STATIC_PROPERTY("Fill", "b", DRAWING_fill),

  GB_STATIC_PROPERTY("Font",  "Font", DRAWING_font),
  GB_STATIC_PROPERTY("Color", "i",    DRAWING_color),

  GB_END_DECLARE
};

