/***************************************************************************

  font.c

  The SDL component

  Font management

  (c) 2004-2005 Laurent Carlier <lordheavy@infonie.fr>
                Benot Minisini <gambas@users.sourceforge.net>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

#define __FONT_C

#include "main.h"
#include "gb_common.h"
#include "font.h"

#include "SDL.h"

/* SFont includes */
#include "SFont.h"
#include "SDL_image.h"

/* TTFont include */
#include "SDL_ttf.h"

#define DEFAULT_TTF_SIZE 20

static void releaseFont(CFONT *cfont)
{
  
  if (cfont->fontType == FONT_TTF)
  {
    if (cfont->id)
      TTF_CloseFont (cfont->id);

    return;
  }

  if (cfont->fontType == FONT_SFONT)
  {
    if (cfont->id)
      SFont_FreeFont(cfont->id);

    return;
  }
}

/**************************************************************************

  Font - a class to manage fonts

***************************************************************************/

BEGIN_METHOD(FONT_new, GB_STRING file; GB_INTEGER type; GB_INTEGER size)

  FONTNAME = GB.FileName(STRING(file), LENGTH(file));
  TTF_Font *tfont;
  SFont_Font *sfont;
  
  if (!GB.ExistFile(FONTNAME))
  {
    GB.Error("File &1 does not exist !", STRING(file));
    return;
  }
    
  if (!VARGOPT(type,0))
  {
    tfont = TTF_OpenFont((char *)FONTNAME , VARGOPT(size, DEFAULT_TTF_SIZE));
      
    if (tfont)
    {
      releaseFont(THIS);
      FONTID = tfont;
      FONTTYPE = FONT_TTF;
      FONTQUAL = 1;
      return;
    }

    sfont = SFont_InitFont(IMG_Load(FONTNAME));
      
    if (sfont)
    {
      releaseFont(THIS);
      FONTID = sfont;
      FONTTYPE = FONT_SFONT;
      return;
    }

    GB.Error("Unknown font type !");
    return;
  }

  switch (VARG(type))
  {
    case FONT_TTF:
    {
      tfont = TTF_OpenFont(FONTNAME, VARGOPT(size, DEFAULT_TTF_SIZE));

      if (!tfont)
      {
        GB.Error("Cannot load font : &1",TTF_GetError());
        return;
      }
        
      releaseFont(THIS);
      FONTID = tfont;
      FONTTYPE = FONT_TTF;
      FONTQUAL = 1;
      return;
    }
    case FONT_SFONT:
    {
      sfont = SFont_InitFont(IMG_Load(FONTNAME));
        
      if (!sfont)
      {
        GB.Error("Cannot load font : &1",SDL_GetError());
        return;
      }

      releaseFont(THIS);
      FONTID = sfont;
      FONTTYPE = FONT_SFONT;
      return;
    }
    default:
      GB.Error("Unknown font type !");
  }

END_METHOD

BEGIN_METHOD_VOID(FONT_free)

  releaseFont(THIS);

END_METHOD

BEGIN_METHOD(FONT_width, GB_STRING text)

  if (FONTTYPE==FONT_TTF)
  {
    int w,h = 0;
    TTF_SizeUTF8(FONTID, STRING(text), &w, &h);
    GB.ReturnInteger(w);
    return;
  }
  if (FONTTYPE==FONT_SFONT)
    GB.ReturnInteger(SFont_TextWidth(FONTID, STRING(text)));

END_METHOD

BEGIN_METHOD(FONT_height, GB_STRING text)

  if (FONTTYPE==FONT_TTF)
  {
    int w,h = 0;
    TTF_SizeUTF8(FONTID, STRING(text), &w, &h);
    GB.ReturnInteger(h);
    return;
  }
  if (FONTTYPE==FONT_SFONT)
    GB.ReturnInteger(SFont_TextHeight(FONTID));

END_METHOD

BEGIN_PROPERTY(FONT_type)

  GB.ReturnInteger(FONTTYPE);

END_PROPERTY

BEGIN_PROPERTY(FONT_ascent)

  if (FONTTYPE==FONT_TTF)
    GB.ReturnInteger(TTF_FontAscent(FONTID));

END_PROPERTY

BEGIN_PROPERTY(FONT_descent)

  if (FONTTYPE==FONT_TTF)
    GB.ReturnInteger(TTF_FontDescent(FONTID));

END_PROPERTY

BEGIN_PROPERTY(FONT_bold)

  if (FONTTYPE==FONT_TTF)
  {
    if (READ_PROPERTY)
      GB.ReturnBoolean((TTF_GetFontStyle(FONTID) & TTF_STYLE_BOLD) ? 1 : 0);
    else
    {
      int flags = TTF_GetFontStyle(FONTID);

      if (VPROP(GB_BOOLEAN))
        flags = flags | TTF_STYLE_BOLD;
      else
      {
        int myflags = 0;

        if (myflags & TTF_STYLE_UNDERLINE)
          myflags = myflags | TTF_STYLE_UNDERLINE;

        if (myflags & TTF_STYLE_ITALIC)
          myflags = myflags | TTF_STYLE_ITALIC;

        flags = myflags;
      }

      TTF_SetFontStyle(FONTID, flags);
    }
  }

END_PROPERTY

BEGIN_PROPERTY(FONT_underline)

  if (FONTTYPE==FONT_TTF)
  {
    if (READ_PROPERTY)
      GB.ReturnBoolean((TTF_GetFontStyle(FONTID) & TTF_STYLE_UNDERLINE) ? 1 : 0);
    else
    {
      int flags = TTF_GetFontStyle(FONTID);

      if (VPROP(GB_BOOLEAN))
        flags = flags | TTF_STYLE_UNDERLINE;
      else
      {
        int myflags = 0;

        if (myflags & TTF_STYLE_BOLD)
          myflags = myflags | TTF_STYLE_BOLD;

        if (myflags & TTF_STYLE_ITALIC)
          myflags = myflags | TTF_STYLE_ITALIC;

        flags = myflags;
      }

      TTF_SetFontStyle(FONTID, flags);
    }
  }

END_PROPERTY

BEGIN_PROPERTY(FONT_italic)

  if (FONTTYPE==FONT_TTF)
  {
    if (READ_PROPERTY)
      GB.ReturnBoolean((TTF_GetFontStyle(FONTID) & TTF_STYLE_ITALIC) ? 1 : 0);
    else
    {
      int flags = TTF_GetFontStyle(FONTID);

      if (VPROP(GB_BOOLEAN))
        flags = flags | TTF_STYLE_ITALIC;
      else
      {
        int myflags = 0;

        if (myflags & TTF_STYLE_UNDERLINE)
          myflags = myflags | TTF_STYLE_UNDERLINE;

        if (myflags & TTF_STYLE_ITALIC)
          myflags = myflags | TTF_STYLE_ITALIC;

        flags = myflags;
      }

      TTF_SetFontStyle(FONTID, flags);
    }
  }

END_PROPERTY

BEGIN_PROPERTY(FONT_size)

  if (READ_PROPERTY)
  {
    if (FONTTYPE == FONT_TTF)
      GB.ReturnInteger(TTF_FontHeight(FONTID)-1);

    if (FONTTYPE == FONT_SFONT)
      GB.ReturnInteger(SFont_TextHeight(FONTID));
      
  }
  else
  {
    if (FONTTYPE == FONT_TTF)
    {
      int style = TTF_GetFontStyle(FONTID);

      TTF_Font *tfont = TTF_OpenFont(FONTNAME, VPROP(GB_INTEGER));

      if (tfont)
      {
        releaseFont(THIS);
        FONTID = tfont;
        TTF_SetFontStyle(FONTID, style);
      }
      else
        GB.Error("Cannot change size : &1",TTF_GetError());
    }
  }

END_PROPERTY

BEGIN_PROPERTY(FONT_aliased)

  if (READ_PROPERTY)
  {
    if (FONTTYPE == FONT_TTF)
      GB.ReturnBoolean(FONTQUAL);
    else
      GB.ReturnBoolean(0);
  }
  else
  {
    if (FONTTYPE == FONT_TTF)
      FONTQUAL = VPROP(GB_BOOLEAN);
  }

END_PROPERTY

/**************************************************************************/

GB_DESC Cfont[] =
{
  GB_DECLARE("Font", sizeof(CFONT)),

  GB_METHOD("_new",  NULL, FONT_new,  "(file)s[(type)i(size)i]"),
  GB_METHOD("_free", NULL, FONT_free, NULL),

  GB_METHOD("Width",  "i", FONT_width,  "(text)s"),
  GB_METHOD("Height", "i", FONT_height, "(text)s"),

  GB_PROPERTY_READ("Type",    "i", FONT_type),
  GB_PROPERTY_READ("Ascent",  "i", FONT_ascent),
  GB_PROPERTY_READ("Descent", "i", FONT_descent),

  GB_PROPERTY("Bold",       "b", FONT_bold),
  GB_PROPERTY("Underline",  "b", FONT_underline),
  GB_PROPERTY("Italic",     "b", FONT_italic),
  GB_PROPERTY("Size",       "i", FONT_size),
  GB_PROPERTY("Aliased",    "b", FONT_aliased),

  GB_CONSTANT("TTF",   "i", FONT_TTF),
  GB_CONSTANT("SFont", "i", FONT_SFONT),

  GB_END_DECLARE
};
