/***************************************************************************

  surface.c

  Gambas extension using SDL

  (c) 2004-2005 Laurent Carlier <lordheavy@infonie.fr>
                Benot Minisini <gambas@users.sourceforge.net>
                Fabien Bodard   <gambasfr@wanadoo.fr>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

#define __SURFACE_C

#include "main.h"
#include "gb_common.h"

#include "surface.h"

/* mirror function include */
#include <string.h>

#include "SDL.h"
#include "SDL_image.h"
#include "SDL_rotozoom.h"

static void create(CSURFACE **surface)
{
  static GB_CLASS class_id = NULL;

  if (!class_id)
    class_id = GB.FindClass("Surface");

  GB.New((void **)surface, class_id, NULL, NULL);
}

/**************************************************************************

  Surface - a class to manage graphics surfaces (aka SDL_surfaces)

***************************************************************************/

BEGIN_METHOD(SURFACE_new, GB_INTEGER Width; GB_INTEGER Height; GB_INTEGER Depth)

  Uint32 myFlags = 0;
  SDL_Surface *surface;

  // enable hardware surface if the driver support it :)
  if (DriverInfo->hw_available)
     myFlags |= SDL_HWSURFACE;
  else
     myFlags |= SDL_SWSURFACE;

  surface = SDL_CreateRGBSurface (myFlags, VARGOPT(Width,0), VARGOPT(Height,0), VARGOPT(Depth,0), 0, 0, 0, 0);

  if (!surface)
  {
    GB.Error("Cannot create Buffer : &1",SDL_GetError());
    return;
  }

  if (SURFACEID)
    SDL_FreeSurface(SURFACEID);

  SURFACEID   = surface;
  THIS->alpha = SURFACEID->format->alpha;
  THIS->key   = SURFACEID->format->colorkey;

END_METHOD

BEGIN_METHOD_VOID(SURFACE_free)

  if (SURFACEID)
    SDL_FreeSurface(SURFACEID);

END_METHOD

BEGIN_METHOD(SURFACE_clear, GB_INTEGER Red; GB_INTEGER Green; GB_INTEGER Blue)

  Uint32 Color;

  Color = SDL_MapRGB(SURFACEID->format, (VARGOPT(Red,0) & 0xFF), (VARGOPT(Green,0) & 0xFF),
         (VARGOPT(Blue,0) & 0xFF));

  if (SDL_FillRect(SURFACEID, NULL, Color)<0)
    GB.Error(SDL_GetError());

END_METHOD

BEGIN_METHOD(SURFACE_blit, GB_INTEGER x; GB_INTEGER y; GB_OBJECT surface; GB_INTEGER srcx; GB_INTEGER srcy; GB_INTEGER srcw; GB_INTEGER srch;)

  CSURFACE *surface;
  SDL_Rect srcRect, destRect;
  SDL_Surface *dest;

  if (MISSING(surface) || (VARG(surface) == NULL))
  {
    if (!MainScreen)
    {
      GB.Error("The Screen is not set !");
      return;
    }
    dest = MainScreen;
  }
  else
  {
    surface = VARG(surface);
    if (GB.CheckObject((void *) surface))
      return;

    dest = GETSURFACE(surface);
  }

  srcRect.x = VARGOPT(srcx,0);
  srcRect.y = VARGOPT(srcy,0);
  srcRect.w = VARGOPT(srcw, (SURFACEID->w-((Uint16) srcRect.x)));
  srcRect.h = VARGOPT(srch, (SURFACEID->h-((Uint16) srcRect.y)));
  destRect.x = VARG(x);
  destRect.y = VARG(y);
  if ((SDL_BlitSurface(SURFACEID, &srcRect, dest, &destRect))<0)
  {
    GB.Error("Cannot blit : &1",SDL_GetError());
    return;
  }

END_METHOD

BEGIN_METHOD(SURFACE_load, GB_STRING file; GB_BOOLEAN convert)

  char *addr;
  long len;

  if (!(GB.LoadFile(STRING(file), LENGTH(file), &addr, &len)))
  {
    int convert = VARGOPT(convert, 1);
    SDL_Surface *surface = IMG_Load_RW(SDL_RWFromMem(addr, len), TRUE);
    GB.ReleaseFile(&addr, len);

    if (!surface)
    {
      GB.Error("Cannot load the Surface : &1",SDL_GetError());
      return;
    }
    if (MainScreen && convert)
      surface = SDL_DisplayFormat(surface);
      
    if (SURFACEID)
      SDL_FreeSurface(SURFACEID);

    SURFACEID = surface;
    THIS->alpha = SURFACEID->format->alpha;
    THIS->key   = SURFACEID->format->colorkey;
  }

END_METHOD

BEGIN_METHOD(SURFACE_convert, GB_OBJECT surface)

  CSURFACE *surface;
  SDL_Surface *tempsurf;

  if (GB.CheckObject(THIS))
    return;

  if (MISSING(surface) || (VARG(surface) == NULL))
  {
    if (MainScreen)
    {
      tempsurf = SDL_ConvertSurface(SURFACEID, MainScreen->format, MainScreen->flags);
      SDL_FreeSurface(SURFACEID);
      SURFACEID = tempsurf;
      THIS->alpha = SURFACEID->format->alpha;
      THIS->key   = SURFACEID->format->colorkey;
    }
    return;
  }
  surface = VARG(surface);

  if (GB.CheckObject((void *) surface))
    return;

  tempsurf = SDL_ConvertSurface(SURFACEID, GETSURFACE(surface)->format, GETSURFACE(surface)->flags);
  SDL_FreeSurface(SURFACEID);
  SURFACEID = tempsurf;
  THIS->alpha = SURFACEID->format->alpha;
  THIS->key   = SURFACEID->format->colorkey;

END_METHOD

BEGIN_METHOD(SURFACE_rotate, GB_INTEGER angle)

  CSURFACE *surface;
  SDL_Surface *tempsurf;

  if (GB.CheckObject(THIS))
    return;

  create(&surface);
  tempsurf = rotozoomSurface (SURFACEID, VARG(angle), 1, 0);
  
  if (tempsurf->format->BitsPerPixel != surface->id->format->BitsPerPixel)
    tempsurf = SDL_ConvertSurface(tempsurf, SURFACEID->format, SURFACEID->flags);
  
  surface->id = tempsurf;
  surface->alpha = tempsurf->format->alpha;
  surface->key = tempsurf->format->colorkey;
  GB.ReturnObject(surface);

END_METHOD

BEGIN_METHOD(SURFACE_stretch, GB_INTEGER width; GB_INTEGER height; GB_BOOLEAN smooth)

  CSURFACE *surface;
  SDL_Surface *tempsurf;
  double zoomx, zoomy;

  if (GB.CheckObject(THIS))
    return;

  zoomx = ((double ) VARG(width))/(SURFACEID->w);
  zoomy = ((double ) VARG(height))/(SURFACEID->h);
  
  create(&surface);
  tempsurf = zoomSurface (SURFACEID, zoomx, zoomy, VARGOPT(smooth,0));
  
  if (tempsurf->format->BitsPerPixel != surface->id->format->BitsPerPixel)
    tempsurf = SDL_ConvertSurface(tempsurf, SURFACEID->format, SURFACEID->flags);
  
  surface->id = tempsurf;
  surface->alpha = tempsurf->format->alpha;
  surface->key = tempsurf->format->colorkey;
  GB.ReturnObject(surface);

END_METHOD

BEGIN_METHOD(SURFACE_zoom, GB_FLOAT zoomx; GB_FLOAT zoomy; GB_BOOLEAN smooth)

  CSURFACE *surface;
  SDL_Surface *tempsurf;

  if (GB.CheckObject(THIS))
    return;

  create(&surface);
  tempsurf = zoomSurface (SURFACEID, VARG(zoomx), VARG(zoomy), VARGOPT(smooth,0));
  
  if (tempsurf->format->BitsPerPixel != surface->id->format->BitsPerPixel)
    tempsurf = SDL_ConvertSurface(tempsurf, SURFACEID->format, SURFACEID->flags);
    
  surface->id = tempsurf;
  surface->alpha = tempsurf->format->alpha;
  surface->key = tempsurf->format->colorkey;
  GB.ReturnObject(surface);

END_METHOD

BEGIN_METHOD_VOID(SURFACE_flip)

  CSURFACE *surface;
  SDL_Surface *flipsurf;
  Uint8 *src, *dest;
  Uint16 scanline;
  Uint16 loop = 0;

  if (GB.CheckObject(THIS))
    return;
    
  flipsurf = SDL_CreateRGBSurface(SURFACEID->flags, SURFACEID->w, SURFACEID->h,
           SURFACEID->format->BitsPerPixel, SURFACEID->format->Rmask, SURFACEID->format->Gmask,
	   SURFACEID->format->Bmask, SURFACEID->format->Amask);
	       
  if (!SDL_MUSTLOCK(SURFACEID))
    SDL_LockSurface(SURFACEID);
    
  if (!SDL_MUSTLOCK(flipsurf))
    SDL_LockSurface(flipsurf);
  
  scanline = SURFACEID->pitch;
  src  = (Uint8 *) SURFACEID->pixels;
  dest = (((Uint8 *)flipsurf->pixels)+((flipsurf->pitch)*(flipsurf->h-1)));
  
  for (loop=0; loop<flipsurf->h; loop++ )
    memcpy (dest-(scanline*loop), src+(scanline*loop), scanline);
  
  if (!SDL_MUSTLOCK(flipsurf))
    SDL_UnlockSurface(flipsurf);
  
  if (!SDL_MUSTLOCK(SURFACEID))
    SDL_UnlockSurface(SURFACEID);
  
  create(&surface);
  surface->id = flipsurf;
  surface->alpha = SURFACEID->format->alpha;
  surface->key = SURFACEID->format->colorkey;
  GB.ReturnObject(surface);

END_METHOD

BEGIN_METHOD_VOID(SURFACE_mirror)

  CSURFACE *surface;
  SDL_Surface *mirrorsurf;
  Uint8 *src, *dest;
  Uint8 BPPixel;
  Uint16 scanline;
  Uint16 loopx, loopy;

  if (GB.CheckObject(THIS))
    return;
    
  mirrorsurf = SDL_CreateRGBSurface(SURFACEID->flags, SURFACEID->w, SURFACEID->h,
               SURFACEID->format->BitsPerPixel, SURFACEID->format->Rmask, SURFACEID->format->Gmask,
	       SURFACEID->format->Bmask, SURFACEID->format->Amask);
	       
  if (!SDL_MUSTLOCK(SURFACEID))
    SDL_LockSurface(SURFACEID);
    
  if (!SDL_MUSTLOCK(mirrorsurf))
    SDL_LockSurface(mirrorsurf);
  
  BPPixel = SURFACEID->format->BytesPerPixel;
  scanline = SURFACEID->pitch;
  src  = (Uint8 *) SURFACEID->pixels;
  dest = (((Uint8 *)mirrorsurf->pixels)+scanline-BPPixel);
  
  for (loopy=0; loopy<(mirrorsurf->h); loopy++)
  {
    for (loopx=0; loopx<(mirrorsurf->w); loopx++)
    {
       memcpy(dest+((scanline*loopy)-(loopx*BPPixel)), src+((scanline*loopy)+(loopx*BPPixel)), BPPixel);
    }
  }  
  
  if (!SDL_MUSTLOCK(mirrorsurf))
    SDL_UnlockSurface(mirrorsurf);
  
  if (!SDL_MUSTLOCK(SURFACEID))
    SDL_UnlockSurface(SURFACEID);
  
  create(&surface);
  surface->id = mirrorsurf;
  surface->alpha = SURFACEID->format->alpha;
  surface->key = SURFACEID->format->colorkey;
  GB.ReturnObject(surface);

END_METHOD

BEGIN_METHOD(SURFACE_overlap, GB_INTEGER x1; GB_INTEGER y1; GB_OBJECT surface2; GB_INTEGER x2; GB_INTEGER y2;)

  CSURFACE *surface2 = VARG(surface2);
  Sint16 x1, x2, y1, y2;
  x1 = VARG(x1);
  x2 = VARG(x2);
  y1 = VARG(y1);
  y2 = VARG(y2);
    
  if (GB.CheckObject(THIS) || GB.CheckObject(surface2))
    return;

  if (!(((x2>=x1) && (x2<(x1+(SURFACEID->w)))) || ((x1>=x2) && (x1<(x2+(GETSURFACE(surface2)->w))))))
  {
    GB.ReturnBoolean(0);
    return;
  }
  
  if (!(((y2>=y1) && (y2<(y1+(SURFACEID->h)))) || ((y1>=y2) && (y1<(y2+(GETSURFACE(surface2)->h))))))
  {
    GB.ReturnBoolean(0);
  }
  else 
    GB.ReturnBoolean(1);

END_METHOD

BEGIN_PROPERTY(SURFACE_width)

  if (SURFACEID)
    GB.ReturnInteger(SURFACEID->w);
  else
    GB.ReturnInteger(0);

END_PROPERTY

BEGIN_PROPERTY(SURFACE_height)

  if (SURFACEID)
    GB.ReturnInteger(SURFACEID->h);
  else
    GB.ReturnInteger(0);

END_PROPERTY

BEGIN_PROPERTY(SURFACE_depth)

  if (SURFACEID)
    GB.ReturnInteger((Uint8 ) SURFACEID->format->BitsPerPixel);
  else
    GB.ReturnInteger(0);

END_PROPERTY

BEGIN_PROPERTY(SURFCHAN_alphavalue)

  Uint32 myflags = 0;

  if (READ_PROPERTY)
    GB.ReturnInteger((Uint8 ) THIS->alpha);
  else
  {
    myflags |= (SDL_SRCALPHA | SDL_RLEACCEL);
    THIS->alpha = (Uint8 ) VPROP(GB_INTEGER);
    if ((SDL_SetAlpha(SURFACEID, myflags, THIS->alpha))<0)
      GB.Error(SDL_GetError());
  }

END_PROPERTY

BEGIN_PROPERTY(SURFCHAN_alphaenabled)

  Uint32 myflags = 0;

  if (READ_PROPERTY)
  {
    if (SURFACEID->flags & SDL_SRCALPHA)
      GB.ReturnBoolean(1);
    else
      GB.ReturnBoolean(0);
  }
  else
  {
    myflags |= SDL_RLEACCEL;

    if (VPROP(GB_BOOLEAN))
      myflags |= SDL_SRCALPHA;

    if ((SDL_SetAlpha(SURFACEID, myflags, THIS->alpha))<0)
      GB.Error(SDL_GetError());
  }

END_PROPERTY

BEGIN_METHOD(SURFCHAN_colorkeyset, GB_INTEGER Red; GB_INTEGER Green; GB_INTEGER Blue)

  Uint8 red, green, blue;
  Uint32 Color, myflags = 0;

  red   = VARG(Red)   & 0xFF;
  green = VARG(Green) & 0xFF;
  blue  = VARG(Blue)  & 0xFF;

  THIS->Ckey.r = red;
  THIS->Ckey.g = green;
  THIS->Ckey.b = blue;

  if (!(SURFACEID->flags & SDL_SRCCOLORKEY))
    return;
  
  myflags |= (SDL_RLEACCEL | SDL_SRCCOLORKEY);
  Color = SDL_MapRGB(SURFACEID->format, THIS->Ckey.r, THIS->Ckey.g, THIS->Ckey.b);

  if ((SDL_SetColorKey(SURFACEID, myflags, Color))<0)
    GB.Error(SDL_GetError());

END_METHOD

BEGIN_PROPERTY(SURFCHAN_colorkeyvalue)


  if (READ_PROPERTY)
    GB.ReturnInteger((Uint32 ) SDL_MapRGB(SURFACEID->format, THIS->Ckey.r, THIS->Ckey.g, THIS->Ckey.b));
  else
  {
    Uint32 myflags = 0;
    Uint32 Color = VPROP(GB_INTEGER);
    THIS->Ckey.r = (Color >> 16) & 0xFF;
    THIS->Ckey.g = (Color >> 8) & 0xFF;
    THIS->Ckey.b = Color & 0xFF;

    myflags |= (SDL_RLEACCEL | SDL_SRCCOLORKEY);
    Color = SDL_MapRGB(SURFACEID->format, THIS->Ckey.r, THIS->Ckey.g, THIS->Ckey.b);

    if ((SDL_SetColorKey(SURFACEID, myflags, Color))<0)
      GB.Error(SDL_GetError());
  }

END_PROPERTY

BEGIN_PROPERTY(SURFCHAN_colorkeyenabled)

  if (READ_PROPERTY)
  {
    if (SURFACEID->flags & SDL_SRCCOLORKEY)
      GB.ReturnBoolean(1);
    else
      GB.ReturnBoolean(0);
  }
  else
  {
    Uint32 Color = 0;
    Uint32 myflags = 0;

    if (VPROP(GB_BOOLEAN))
      myflags |= SDL_SRCCOLORKEY;
    
    myflags |= SDL_RLEACCEL;

    Color = SDL_MapRGB(SURFACEID->format, THIS->Ckey.r, THIS->Ckey.g, THIS->Ckey.b);

    if ((SDL_SetColorKey(SURFACEID, myflags, Color))<0)
      GB.Error(SDL_GetError());
  }

END_PROPERTY

/**************************************************************************/

GB_DESC Csurfacealpha[] =
{
  GB_DECLARE(".Alpha",0), GB_VIRTUAL_CLASS(),

  GB_PROPERTY("Value",   "i", SURFCHAN_alphavalue),
  GB_PROPERTY("Enabled", "b", SURFCHAN_alphaenabled),

  GB_END_DECLARE
};

GB_DESC Csurfacecolor[] =
{
  GB_DECLARE(".ColorKey",0), GB_VIRTUAL_CLASS(),

  GB_METHOD("Set", NULL, SURFCHAN_colorkeyset, "(Red)i(Green)i(Blue)i"),

  GB_PROPERTY("Value",   "i", SURFCHAN_colorkeyvalue),
  GB_PROPERTY("Enabled", "b", SURFCHAN_colorkeyenabled),

  GB_END_DECLARE
};

GB_DESC Csurface[] =
{
  GB_DECLARE("Surface", sizeof(CSURFACE)),

  GB_METHOD("_new",  NULL, SURFACE_new,  "[(Width)i(Height)i(Depth)i]"),
  GB_METHOD("_free", NULL, SURFACE_free, NULL),

  GB_METHOD("Clear",   NULL, SURFACE_clear,   "[(red)i(green)i(blue)i]"),
  GB_METHOD("Blit",    NULL, SURFACE_blit,    "(X)i(Y)i[(Surface)Surface;(SrcX)i(SrcY)i(SrcWidth)i(SrcHeight)i]"),
  GB_METHOD("Load",    NULL, SURFACE_load,    "(file)s[(convert)b]"),
  GB_METHOD("Convert", NULL, SURFACE_convert, "[(Surface)Surface;]"),
  GB_METHOD("Overlap", "b" , SURFACE_overlap, "(X1)i(Y1)i(Surface)Surface;(X2)i(Y2)i"),
  
  GB_METHOD("Rotate",  "Surface", SURFACE_rotate,  "(angle)i"),
  GB_METHOD("Stretch", "Surface", SURFACE_stretch, "(Width)i(Height)i[(Smooth)b]"),
  GB_METHOD("Zoom",    "Surface", SURFACE_zoom,    "(ZoomX)f(ZoomY)f[(Smooth)b]"),
  GB_METHOD("Scale",   "Surface", SURFACE_zoom,    "(ZoomX)f(ZoomY)f[(Smooth)b]"),
  GB_METHOD("Flip",    "Surface", SURFACE_flip,    NULL),  
  GB_METHOD("Mirror",  "Surface", SURFACE_mirror,  NULL),

  GB_PROPERTY_READ("Width",  "i", SURFACE_width),
  GB_PROPERTY_READ("W",      "i", SURFACE_width),
  GB_PROPERTY_READ("Height", "i", SURFACE_height),
  GB_PROPERTY_READ("H",      "i", SURFACE_height),
  GB_PROPERTY_READ("Depth",  "i", SURFACE_depth),
  GB_PROPERTY_READ("D",      "i", SURFACE_depth),
  
  GB_PROPERTY_SELF("Alpha",    ".Alpha"),
  GB_PROPERTY_SELF("ColorKey", ".ColorKey"),

  GB_END_DECLARE
};
