/***************************************************************************

  archive.c

  The archive management routines

  (c) 2000-2005 Benot Minisini <gambas@users.sourceforge.net>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

#define __GBX_ARCHIVE_C

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "gb_common.h"
#include "gb_error.h"
#include "gbx_value.h"
#include "gb_limit.h"
#include "gb_alloc.h"
#include "gbx_string.h"
#include "gbx_regexp.h"
#include "gbx_exec.h"
#include "gbx_list.h"
#include "gbx_project.h"
#include "gbx_component.h"
#include "gbx_stream.h"

#include "gbx_archive.h"

/* main archive project (used only if gbx is run with -x flag) */
PUBLIC ARCHIVE *ARCHIVE_main = NULL;

static char *arch_pattern = NULL;
static long arch_index = 0;
static ARCHIVE *arch_dir = NULL;


static void arch_error(const char *msg)
{
  if (msg == NULL)
    THROW(E_ARCH, strerror(errno));
  else
    THROW(E_ARCH, msg);
}

static void set_pos(ARCHIVE *arch, long pos)
{
  if (lseek(arch->fd, pos, SEEK_SET) < 0)
    arch_error(NULL);
}

static void read_at(ARCHIVE *arch, long pos, void *buf, long len)
{
  set_pos(arch, pos);

  if (read(arch->fd, buf, len) != len)
    arch_error(NULL);
}


PUBLIC ARCHIVE *ARCHIVE_create(const char *name)
{
  ARCHIVE *arch;
  char *domain;
  
  ALLOC_ZERO(&arch, sizeof(ARCHIVE), "ARCH_create");
  
  arch->name = name;
  
  TABLE_create(&arch->classes, sizeof(CLASS_SYMBOL), TF_IGNORE_CASE);
  
  domain = FILE_buffer();
  
  if (name)
    STRING_new(&arch->domain, name, 0);
  else
    STRING_new(&arch->domain, "gb", 0);
  
  arch->translation_loaded = FALSE;
  
  return arch;
}

static void load_exported_class(ARCHIVE *arch)
{
  /*STREAM stream;*/
  char *buffer;
  long len;
  char *name;
  
  /* COMPONENT_current is set => it will look in the archive */
  
  STREAM_load(".list", &buffer, &len);
  
  /*STREAM_open(&stream, ".list", ST_READ);
  STREAM_lof(&stream, &len);

  ALLOC(&buffer, len + 1, "load_exported_class");

  STREAM_read(&stream, buffer, len);
  STREAM_close(&stream);*/

  /* The file must end with a newline !*/
  buffer[len - 1] = 0;

  #if DEBUG_COMP
    fprintf(stderr, ".list file:\n-----------\n\n%s\n----------\n\n", buffer);
  #endif
  
  name = strtok(buffer, "\n");
  while (name)
  {
    #if DEBUG_COMP
      fprintf(stderr, "Load %s global\n", name);
    #endif
  
    CLASS_check_global(name);
    CLASS_load(CLASS_find_global(name));
    name = strtok(NULL, "\n");
  }
  
  FREE(&buffer, "load_exported_class");
}

static void load_archive(ARCHIVE *arch, const char *path)
{
  long len;
  int i;
  long pos;
  
  arch->fd = open(path, O_RDONLY);
  if (arch->fd < 0)
    THROW(E_OPEN, path, strerror(errno));

  /* Lecture de l'entte */

  read_at(arch, 32, &arch->header, sizeof(ARCH_HEADER));

  /* Lecture des chanes */

  len = arch->header.pos_table - arch->header.pos_string;
  if (len <= 0)
    arch_error("corrupted header");

  ALLOC(&arch->string, len, "ARCH_init");
  read_at(arch, arch->header.pos_string, arch->string, len);

  /* Lecture de la table */

  len = arch->header.n_symbol * sizeof(ARCH_SYMBOL);
  if (len <= 0)
    arch_error("corrupted header");

  ALLOC(&arch->symbol, len, "ARCHIVE_load");
  read_at(arch, arch->header.pos_table, arch->symbol, len);

  /* Relocation des chanes */

  pos = 0;
  for (i = 0; i < arch->header.n_symbol; i++)
  {
    arch->symbol[i].sym.name = &arch->string[pos];
    pos += arch->symbol[i].sym.len;
  } 

  if (arch != ARCHIVE_main)  
    load_exported_class(arch);
}


PUBLIC void ARCHIVE_load(ARCHIVE *arch)
{
  char *path = FILE_buffer();
  
  sprintf(path, ARCH_PATTERN, COMPONENT_path, arch->name);
  load_archive(arch, path);
}


PUBLIC void ARCHIVE_create_main(const char *path)
{
  ARCHIVE_main = ARCHIVE_create(NULL);
  load_archive(ARCHIVE_main, path);
}


PUBLIC void ARCHIVE_delete(ARCHIVE *arch)
{
  if (arch->fd)
  {
    FREE(&arch->string, "ARCHIVE_delete");
    FREE(&arch->symbol, "ARCHIVE_delete");
    close(arch->fd);
  }

  TABLE_delete(&arch->classes);
  STRING_free(&arch->domain);
  
  FREE(&arch, "ARCHIVE_delete");
}


PUBLIC void ARCHIVE_init(void)
{
  //ARCH_create(path);
  //ARCHIVE_main = _arch_list;
}


PUBLIC void ARCHIVE_exit(void)
{
  if (ARCHIVE_main)
    ARCHIVE_delete(ARCHIVE_main);
  
  STRING_free(&arch_pattern);
}

#if 0
PUBLIC void ARCH_add(const char *path)
{
  const char *arch_path;
  struct stat info;
  char *name;

  if (FILE_is_relative(path))
  {
    arch_path = FILE_cat(FILE_get_dir(PROJECT_path), path, NULL);
    if (stat(arch_path, &info) == 0)
    {
      if (S_ISDIR(info.st_mode))
      {
        STRING_new(&name, FILE_get_name(arch_path), 0);
        arch_path = FILE_cat(arch_path, name, NULL);
        STRING_free(&name);
      }
    }
    else
      arch_path = FILE_cat(PROJECT_exec_path, "bin", path, NULL);
  }
  else
    arch_path = path;

  if (stat(arch_path, &info) != 0)
    THROW(E_LIBRARY, path, "File not found");

  ARCH_create(arch_path);
}
#endif

PUBLIC bool ARCHIVE_get_current(ARCHIVE **parch)
{
  if (*parch)
    return FALSE;
    
  if (COMPONENT_current && COMPONENT_current->archive)
    *parch = COMPONENT_current->archive;
  else if (CP && CP->component && CP->component->archive)
    *parch = CP->component->archive;
  else
    *parch = ARCHIVE_main;

  return *parch == NULL;
}

PUBLIC bool ARCHIVE_get(ARCHIVE *arch, const char *path, int len_path, ARCH_FIND *find)
{
  long index;
  ARCH_SYMBOL *sym;

  if (len_path <= 0)
    len_path = strlen(path);

  if (len_path == 0)
    return TRUE;
    
  if (ARCHIVE_get_current(&arch))
    return TRUE;

  SYMBOL_find(arch->symbol, arch->header.n_symbol, sizeof(ARCH_SYMBOL), TF_NORMAL, path, len_path, 0, &index);
  if (index == NO_SYMBOL)
    return TRUE;

  sym = &arch->symbol[index];

  find->sym = sym;
  find->pos = find->sym->pos;
  find->len = sym->len;
  find->arch = arch;

  //ARCHIVE_last = arch;

  return FALSE;
}


PUBLIC bool ARCHIVE_exist(ARCHIVE *arch, const char *path)
{
  ARCH_FIND find;

  if (ARCHIVE_get_current(&arch))
    return FALSE;
    
  return (!ARCHIVE_get(arch, path, 0, &find));
}


PUBLIC void ARCHIVE_stat(ARCHIVE *arch, const char *path, FILE_STAT *info)
{
  ARCH_FIND find;
  struct stat buf;

  if (ARCHIVE_get_current(&arch))
    THROW_SYSTEM(ENOENT, path);
    
  if (!ARCHIVE_get(arch, path, 0, &find))
    THROW_SYSTEM(ENOENT, path);

  fstat(find.arch->fd, &buf);

  info->type = GB_STAT_FILE;
  info->mode = 0400;

  info->size = find.len;
  info->atime = (long)buf.st_mtime;
  info->mtime = (long)buf.st_mtime;
  info->ctime = (long)buf.st_mtime;
  info->hidden = (*FILE_get_name(path) == '.');
  info->uid = buf.st_uid;
  info->gid = buf.st_gid;
}


PUBLIC bool ARCHIVE_read(ARCHIVE *arch, long pos, void *buffer, long len)
{
  if (lseek(arch->fd, pos, SEEK_SET) < 0)
    return TRUE;

  if (read(arch->fd, buffer, len) != len)
    return TRUE;

  return FALSE;
}


PUBLIC void ARCHIVE_dir_first(ARCHIVE *arch, const char *path, const char *pattern)
{
  if (pattern == NULL)
    pattern = "*";

  if (ARCHIVE_get_current(&arch))
  {
    arch_dir = NULL;
    return;
  }
    
  path = FILE_cat(path, pattern, NULL);

  if (arch_pattern)
    STRING_free(&arch_pattern);
  STRING_new(&arch_pattern, path, FILE_buffer_length());

  arch_index = 0;
  arch_dir = arch;
}


PUBLIC bool ARCHIVE_dir_next(char **name, long *len)
{
  SYMBOL *sym;

  /*if (arch_fd < 0)
    return FILE_dir_next(name, len);*/

  if (!arch_dir)
    return TRUE;

  for(;;)
  {
    if (arch_index >= arch_dir->header.n_symbol)
      return TRUE;

    sym = (SYMBOL *)&arch_dir->symbol[arch_index];
    sym = (SYMBOL *)&arch_dir->symbol[sym->sort];

    if (arch_pattern == NULL)
      break;

    if (REGEXP_match(arch_pattern, STRING_length(arch_pattern), sym->name, sym->len))
      break;

    arch_index++;
  }

  *name = sym->name;
  *len = sym->len;
  return FALSE;
}

