/***************************************************************************

  component.c

  GAMBAS component management routines

  (c) 2000-2005 Benot Minisini <gambas@users.sourceforge.net>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

#define __GBX_COMPONENT_C

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "gb_common.h"
#include "gb_common_buffer.h"
#include "gb_common_case.h"
#include "gb_error.h"
#include "gb_alloc.h"
#include "gbx_list.h"
#include "gb_replace.h"
#include "gb_file.h"
#include "gbx_debug.h"

#include <fcntl.h>
#include <stdlib.h>
#include <stdarg.h>
#include <time.h>
#include <ctype.h>

#include "gb_error.h"

#include "gbx_class.h"
#include "gbx_exec.h"

#include "gbx_local.h"
#include "gbx_archive.h"
#include "gbx_library.h"

#include "gbx_component.h"


/*#define DEBUG*/
/*#define DEBUG_PRELOAD*/

PUBLIC COMPONENT *COMPONENT_current = NULL;
PUBLIC COMPONENT *COMPONENT_main;
PUBLIC int COMPONENT_count = 0;
PUBLIC char *COMPONENT_path;

static COMPONENT *_component_list = NULL;


PUBLIC void COMPONENT_init(void)
{
  LIBRARY_init();
  ARCHIVE_init();
}


PUBLIC void COMPONENT_exit(void)
{
  COMPONENT *comp;

  LIST_for_each(comp, _component_list)
    COMPONENT_unload(comp);

  while (_component_list)
    COMPONENT_delete(_component_list);

  LIBRARY_exit();
  ARCHIVE_exit();    
  
  STRING_free(&COMPONENT_path);
}



PUBLIC void COMPONENT_load_all(void)
{
  COMPONENT *comp;

  if (EXEC_debug)
    COMPONENT_create("gb.eval");

  LIST_for_each(comp, _component_list)
  {
    comp->preload = TRUE;
    COMPONENT_load(comp);
  }
}


PUBLIC COMPONENT *COMPONENT_find(const char *name)
{
  COMPONENT *comp;

  LIST_for_each(comp, _component_list)
  {
    if (strcmp(comp->name, name) == 0)
      return comp;
  }

  return NULL;
}


PUBLIC COMPONENT *COMPONENT_create(const char *name)
{
  COMPONENT *comp;
  char *path;

  comp = COMPONENT_find(name);
  if (comp)
    return comp;

  ALLOC_ZERO(&comp, sizeof(COMPONENT), "COMPONENT_create");

  comp->class = CLASS_Component;
  comp->ref = 1;

  STRING_new(&comp->name, name, 0);
  
  path = FILE_buffer();
  sprintf(path, LIB_PATTERN, COMPONENT_path, name);
    
  if (FILE_exist(path))
    comp->library = LIBRARY_create(name);
  else
  {
    path = FILE_buffer();
    sprintf(path, ARCH_PATTERN, COMPONENT_path, name);
    
    if (FILE_exist(path))
      comp->archive = ARCHIVE_create(name);
    else
      THROW(E_LIBRARY, comp->name, "cannot find library file");
  }
  
  LIST_insert((void **)&_component_list, comp, &comp->list);
  COMPONENT_count++;

  return comp;
}


PUBLIC void COMPONENT_delete(COMPONENT *comp)
{
  COMPONENT_unload(comp);
  LIST_remove((void **)&_component_list, comp, &comp->list);
  COMPONENT_count--;

  if (comp->library)
    LIBRARY_delete(comp->library);
  else if (comp->archive)
    ARCHIVE_delete(comp->archive);
  
  STRING_free(&comp->name);

  FREE(&comp, "COMPONENT_delete");
}


PUBLIC void COMPONENT_load(COMPONENT *comp)
{
  COMPONENT *current;
  
  if (comp->loaded)
    return;

  #if DEBUG_COMP
    fprintf(stderr, "Loading component %s\n", comp->name);
  #endif
      
  current = COMPONENT_current;  
  COMPONENT_current = comp;
  
  if (comp->library)
    LIBRARY_load(comp->library);
  else if (comp->archive)
    ARCHIVE_load(comp->archive);
    
  comp->loaded = TRUE;
  COMPONENT_current = current;
}


PUBLIC void COMPONENT_unload(COMPONENT *comp)
{
  if (comp->library)
    LIBRARY_unload(comp->library);
}


PUBLIC COMPONENT *COMPONENT_next(COMPONENT *comp)
{
  if (comp)
    return (COMPONENT *)(comp->list.next);
  else
    return _component_list;
}


PUBLIC void COMPONENT_translation_must_be_reloaded(void)
{
  COMPONENT *comp;

  LIST_for_each(comp, _component_list)
  {
    if (comp->archive)
      comp->archive->translation_loaded = FALSE;
  }
}


#if 0
PUBLIC const char *LIBRARY_get_control_list(LIBRARY *lib)
{
  void (*gambas_info)();
  const char *result = NULL;

  gambas_info = lt_dlsym(lib->handle, LIB_INFO);
  if (gambas_info == NULL)
    return NULL;

  (*gambas_info)(GB_INFO_CONTROL, &result);
  return result;
}
#endif

